/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

#ifndef __ORCUS_OOXML_TOKENS_HPP__
#define __ORCUS_OOXML_TOKENS_HPP__

#include "orcus/tokens.hpp"

namespace orcus {

extern tokens ooxml_tokens;
extern tokens opc_tokens;

}

#endif
/* vim:set shiftwidth=4 softtabstop=4 expandtab: */
